#include "localsolver.h"
#include <fstream>
#include <iostream>
#include <string.h>
#include <vector>

using namespace localsolver;
using namespace std;

class CapacitatedFacilityLocation {
public:
    // Data from the problem
    int nbMaxFacilities;
    int nbSites;
    vector<double> capacityData;
    vector<double> openingPriceData;
    vector<double> demandData;
    vector<vector<double>> allocationPriceData;

    // LocalSolver
    LocalSolver localsolver;

    // Variables
    vector<LSExpression> facilityAssignments;

    // Objective
    vector<LSExpression> cost;
    LSExpression totalCost;

    void solve(int limit) {
        // Declare the optimization model
        LSModel model = localsolver.getModel();

        // Table of sets representing the sites provided by each facility
        facilityAssignments.resize(nbMaxFacilities);
        for (int f = 0; f < nbMaxFacilities; ++f) {
            facilityAssignments[f] = model.setVar(nbSites);
        }

        // Each site is covered by exactly one facility
        model.constraint(model.partition(facilityAssignments.begin(), facilityAssignments.end()));

        // Converting demand and allocationPrice into LocalSolver array
        LSExpression demand = model.array(demandData.begin(), demandData.end());
        LSExpression allocationPrice = model.array();
        for (int f = 0; f < nbMaxFacilities; ++f) {
            allocationPrice.addOperand(model.array(allocationPriceData[f].begin(), allocationPriceData[f].end()));
        }

        cost.resize(nbMaxFacilities);
        for (int f = 0; f < nbMaxFacilities; ++f) {
            LSExpression facility = facilityAssignments[f];
            LSExpression size = model.count(facility);

            // Capacity constraint
            LSExpression demandLambda = model.createLambdaFunction([&](LSExpression i) { return demand[i]; });
            model.constraint(model.sum(facility, demandLambda) <= capacityData[f]);

            // Cost (allocation price + opening price)
            LSExpression costLambda =
                model.createLambdaFunction([&](LSExpression i) { return model.at(allocationPrice, f, i); });
            cost[f] = model.sum(facility, costLambda) + (size > 0) * openingPriceData[f];
        }

        // Objective : minimize total cost
        totalCost = model.sum(cost.begin(), cost.end());
        model.minimize(totalCost);
        model.close();

        // Parametrize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /* Read instance data */
    void readInstance(const string& fileName) {
        ifstream infile;
        infile.exceptions(ifstream::failbit | ifstream::badbit);
        infile.open(fileName.c_str());

        infile >> nbMaxFacilities;
        infile >> nbSites;
        capacityData.resize(nbMaxFacilities);
        openingPriceData.resize(nbMaxFacilities);
        demandData.resize(nbSites);
        allocationPriceData.resize(nbMaxFacilities, vector<double>(nbSites));

        for (int f = 0; f < nbMaxFacilities; ++f) {
            // List of facilities capacities
            infile >> capacityData[f];
            // List of fixed costs induced by the facilities opening
            infile >> openingPriceData[f];
        }

        // Demand of each site
        for (int s = 0; s < nbSites; ++s) {
            infile >> demandData[s];
        }

        // Allocation price between sites and facilities
        for (int f = 0; f < nbMaxFacilities; ++f) {
            for (int s = 0; s < nbSites; ++s) {
                infile >> allocationPriceData[f][s];
            }
        }

        infile.close();
    }

    /* Write the solution in a file with the following format:
     * - value of the objective
     * - indices of the open facilities followed by all the sites they provide */
    void writeSolution(const string& fileName) {
        ofstream outfile;
        outfile.exceptions(ofstream::failbit | ofstream::badbit);
        outfile.open(fileName.c_str());
        outfile << totalCost.getDoubleValue() << endl;
        for (int f = 0; f < nbMaxFacilities; ++f) {
            if (cost[f].getDoubleValue() > 0) {
                outfile << f << endl;
                LSCollection sites_assigned = facilityAssignments[f].getCollectionValue();
                for (lsint s = 0; s < sites_assigned.count(); ++s) {
                    outfile << sites_assigned[s] << " ";
                }
                outfile << endl;
            }
        }
    }
};

int main(int argc, char** argv) {
    if (argc < 2) {
        cerr << "Usage: capacitated_facility_location inputFile [outputfile] [timeLimit]" << endl;
        return 1;
    }
    const char* instanceFile = argv[1];
    const char* solFile = argc > 2 ? argv[2] : NULL;
    const char* strTimeLimit = argc > 3 ? argv[3] : "20";
    try {
        CapacitatedFacilityLocation model;
        model.readInstance(instanceFile);
        model.solve(atoi(strTimeLimit));
        if (solFile != NULL)
            model.writeSolution(solFile);
    } catch (const exception& e) {
        cerr << "An error occured " << e.what() << endl;
        return -1;
    }
}